#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pkg_resources import get_distribution
from magma_var import const
from magma_var import earthquakes
from magma_var.magma_var import MagmaVar
from magma_var.download import Download
from magma_var.plot import Plot
from magma_var import utils

__version__ = get_distribution("magma-var").version
__author__ = "Martanto"
__author_email__ = "martanto@live.com"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2024, MAGMA Indonesia"
__url__ = "https://github.com/martanto/magma-var"

__all__ = [
    "__version__",
    "__author__",
    "__author_email__",
    "__license__",
    "__copyright__",
    "const",
    "earthquakes",
    "Plot",
    "MagmaVar",
    "Download",
    "utils",
]



from typing import List, Dict


URL_FILTER = "https://magma.esdm.go.id/api/v1/magma-var/filter"
URL_VOLCANO = "https://magma.esdm.go.id/api/v1/home/gunung-api"
URL_EVALUATION = "https://magma.esdm.go.id/api/v1/python/magma-var/evaluasi"

# Source Excel:
# https://1drv.ms/x/c/ef270b0f3925154b/EbVq29Kvwj9ArjOGIYM5vlIB8zFRIv5THo-r5aWcUuoVnQ?e=ekXAYA
EARTHQUAKE_TYPES: List[Dict[str, str]] = [
    {
        "code": "lts",
        "alias": "Eruption",
        "earthquake_id": "Letusan/Erupsi",
        "earthquake_en": "Eruption",
        "color": "#F44336",
        "column_name": "gempa.letusan_erupsi",
    },
    {
        "code": "apl",
        "alias": "PF - Fountain Collapse",
        "earthquake_id": "Awan Panas Letusan",
        "earthquake_en": "Fountain Collapse Pyroclastic Flow",
        "color": "#E91E63",
        "column_name": "gempa.awan_panas_letusan",
    },
    {
        "code": "apg",
        "alias": "PDC/PF",
        "earthquake_id": "Awan Panas Guguran",
        "earthquake_en": "Pyroclastic Density Current (PDC)",
        "color": "#673AB7",
        "column_name": "gempa.awan_panas_guguran",
    },
    {
        "code": "gug",
        "alias": "Rockfall",
        "earthquake_id": "Guguran",
        "earthquake_en": "Rockfall",
        "color": "#1976D2",
        "column_name": "gempa.guguran",
    },
    {
        "code": "hbs",
        "alias": "Degassing",
        "earthquake_id": "Hembusan",
        "earthquake_en": "Degassing",
        "color": "#3F51B5",
        "column_name": "gempa.hembusan",
    },
    {
        "code": "hrm",
        "alias": "Harmonic",
        "earthquake_id": "Harmonik",
        "earthquake_en": "Harmonic",
        "color": "#607D8B",
        "column_name": "gempa.harmonik",
    },
    {
        "code": "tre",
        "alias": "Non-Harmonic Tremor",
        "earthquake_id": "Tremor Non-Harmonik",
        "earthquake_en": "Non-Harmonic Tremor",
        "color": "#0D47A1",
        "column_name": "gempa.tremor_non-_harmonik",
    },
    {
        "code": "tor",
        "alias": "Tornillo",
        "earthquake_id": "Tornillo",
        "earthquake_en": "Tornillo",
        "color": "#03A9F4",
        "column_name": "gempa.tornillo",
    },
    {
        "code": "lof",
        "alias": "LF",
        "earthquake_id": "Low Frequency",
        "earthquake_en": "Low Frequency",
        "color": "#006064",
        "column_name": "gempa.low_frequency",
    },
    {
        "code": "hyb",
        "alias": "Hybrid/MP",
        "earthquake_id": "Hybrid/Fase Banyak",
        "earthquake_en": "Hybrid/Multi Phase",
        "color": "#009688",
        "column_name": "gempa.hybrid_fase_banyak",
    },
    {
        "code": "vtb",
        "alias": "VT-B",
        "earthquake_id": "Vulkanik Dangkal",
        "earthquake_en": "Shallow Volcanic-Tectonic (VT-B)",
        "color": "#8BC34A",
        "column_name": "gempa.vulkanik_dangkal",
    },
    {
        "code": "vta",
        "alias": "VT-A",
        "earthquake_id": "Vulkanik Dalam",
        "earthquake_en": "Deep Volcanic-Tectonic (VT-A)",
        "color": "#33691E",
        "column_name": "gempa.vulkanik_dalam",
    },
    {
        "code": "vlp",
        "alias": "VLP",
        "earthquake_id": "Very Long Period",
        "earthquake_en": "Very Long Period",
        "color": "#827717",
        "column_name": "gempa.very_long_period",
    },
    {
        "code": "tel",
        "alias": "Local Tectonic",
        "earthquake_id": "Tektonik Lokal",
        "earthquake_en": "Local Tectonic",
        "color": "#F57F17",
        "column_name": "gempa.tektonik_lokal",
    },
    {
        "code": "trs",
        "alias": "Felt Earthquake",
        "earthquake_id": "Terasa",
        "earthquake_en": "Felt Earthquake",
        "color": "#FFCA28",
        "column_name": "gempa.terasa",
    },
    {
        "code": "tej",
        "alias": "Teleseismic",
        "earthquake_id": "Tektonik Jauh",
        "earthquake_en": "Teleseismic",
        "color": "#FFA726",
        "column_name": "gempa.tektonik_jauh",
    },
    {
        "code": "dev",
        "alias": "Double Event",
        "earthquake_id": "Double Event",
        "earthquake_en": "Double Event",
        "color": "#FF5722",
        "column_name": "gempa.double_event",
    },
    {
        "code": "gtb",
        "alias": "Lahar",
        "earthquake_id": "Getaran Banjir",
        "earthquake_en": "Lahar",
        "color": "#795548",
        "column_name": "gempa.getaran_banjir",
    },
    {
        "code": "dpt",
        "alias": "Deep Tremor",
        "earthquake_id": "Deep Tremor",
        "earthquake_en": "Deep Tremor",
        "color": "#333333",
        "column_name": "gempa.deep_tremor",
    },
    {
        "code": "mtr",
        "alias": "Tremor",
        "earthquake_id": "Tremor Menerus",
        "earthquake_en": "Tremor (continuous)",
        "color": "#9E9E9E",
        "column_name": "gempa.tremor_menerus",
    },
]



import json
import os
from datetime import datetime, timezone
from typing import Self

import pandas as pd
import requests
from magma_auth import auth
from magma_var import MagmaVar
from magma_var.const import URL_VOLCANO, URL_FILTER
from magma_var.utils import save, transform, extract


class Download(MagmaVar):
    def __init__(
        self,
        volcano_code: str,
        start_date: str,
        end_date: str,
        token: str = None,
        current_dir: str = None,
        locale: str = "id",
        url_filter: str = None,
        url_volcano: str = None,
        verbose: bool = False,
        debug: bool = False,
    ):
        super().__init__(
            volcano_code, start_date, end_date, current_dir, locale, verbose
        )

        self.token = token
        self.validate()

        self.url_filter = url_filter or URL_FILTER
        self.url_volcano = url_volcano or URL_VOLCANO

        os.makedirs(self.json_dir, exist_ok=True)

        self.page_dir: str = os.path.join(self.json_dir, "pages", self.volcano_code)
        self.daily_dir: str = os.path.join(self.json_dir, "daily", self.volcano_code)

        self.json_filename = f"{self.filename}"
        self.files: list[str] = []
        self.events: list[dict[str, str]] = []
        self._df: pd.DataFrame = pd.DataFrame()

        self.headers = {
            "Authorization": "Bearer " + self.token,
            "Content-Type": "application/json",
        }

        if self.verbose:
            print(f"ℹ️ JSON Directory: {self.json_dir}")
            print(f"ℹ️ Volcano Directory: {self.page_dir}")

        self.debug = debug

    def validate(self) -> None:
        """Validating parameters.

        Returns:
            None

        Raise:
            ValueError: If token not found or expired.
        """
        if self.token is None:
            success, token, _expired_at = auth.load_token()
            expired_at = datetime.fromisoformat(token["expires_at"])
            now = datetime.now(timezone.utc)

            assert success, ValueError(f"❌ Token not found. Please check your token.")

            assert now > expired_at, ValueError(
                f"❌ Token is expired. Expiration time: {expired_at}"
            )

            if self.verbose:
                print(f"ℹ️ Token: {token}")
                print(f"ℹ️ Expiration Date: {expired_at}")
            self.token = token

    @property
    def df(self) -> pd.DataFrame:
        return self._df

    @df.setter
    def df(self, values: list[dict[str, str]]):
        self._df = pd.DataFrame(values)

    def _download_daily(self, data: dict, overwrite: bool = False) -> None:
        """Saving daily json file

        Args:
            data (dict): Data from response['data']
            overwrite (bool, optional): Overwrite existing file. Defaults to False.

        Returns:
            None
        """
        notice_number: str = data["laporan_terakhir"]["noticenumber"]
        year = notice_number[0:4]
        month = notice_number[4:6]
        day = notice_number[6:8]
        filename = f"{year}-{month}-{day}.json"
        filepath = os.path.join(str(self.daily_dir), filename)

        if os.path.isfile(filepath) and not overwrite:
            if self.verbose:
                print(f"ℹ️ File Exists. Skipping: {filepath}")
            return None

        save(filepath, data)

        if self.verbose:
            print(f"🗃️ File Saved: {filepath}")

        return None

    def download_daily(self, overwrite: bool = False) -> None:
        """Download as daily json file

        Args:
            overwrite (bool, optional): Overwrite file if exists. Defaults to False.
        """
        os.makedirs(self.daily_dir, exist_ok=True)

        if self.verbose:
            print("=" * 60)
            print(f"ℹ️ Downloading first page")

        response = self.download()
        first_page_data = response["data"]

        if self.verbose:
            print(f"ℹ️ Data found : {len(first_page_data)}")
            print("=" * 60)

        for _data in first_page_data:
            self._download_daily(_data, overwrite)

        # Downloading rest of data
        last_page = response["last_page"]
        if last_page == 1:
            return None

        if self.verbose:
            print("=" * 60)
            print(f"ℹ️ Downloading from page 2 to {last_page}")
            print("=" * 60)

        for page in range(2, last_page + 1):
            response = self.download({"page": page})
            for _data in response["data"]:
                self._download_daily(_data, overwrite)

        return None

    def download(self, params: dict = None) -> dict:
        """Download VAR from Filter URL

        Args:
            params (dict): Page parameter to pass

        Returns:
            response (dict): Dictionary of response
        """
        volcano_code = self.volcano_code
        start_date = self.start_date
        end_date = self.end_date

        payload: str = json.dumps(
            {
                "code": volcano_code,
                "start_date": start_date,
                "end_date": end_date,
            }
        )

        try:
            response = requests.request(
                "GET",
                self.url_filter,
                headers=self.headers,
                data=payload,
                params=params,
            ).json()

            if "errors" in response.keys():
                raise ValueError(f'❌ Download Error :: {response["errors"]}')

            if "data" not in response.keys():
                raise KeyError(f"❌ Data not found in response :: {response}")

            if len(response["data"]) == 0:
                raise ValueError(
                    f"❌ No data found for {self.volcano_code} :: {response}"
                )

            return response
        except Exception as e:
            raise ValueError(f"❌ Failed to download JSON :: {e}. {params}")

    def download_per_page(self, response: dict) -> list[str]:
        """Download VAR per page from Filter URL

        Args:
            response (dict): Dictionary of response

        Returns:
            list of filenames
        """
        pages: list[str] = []
        last_page = response["last_page"]

        if last_page == 1:
            pages.append(str(last_page))
            return pages

        if self.verbose:
            print(f"ℹ️ Downloading from page 2 to {last_page}")
            print("=" * 60)

        # Starting download from page 2 to last page
        for page in range(2, last_page + 1):
            # code_startDate_endDate_page.json
            # example: AWU_2025-01-01_2025-01-31_1.json
            filename = f"{self.json_filename}_{page}.json"
            json_per_page = os.path.join(self.page_dir, filename)

            if os.path.exists(json_per_page):
                if self.verbose:
                    print(f"✅ Skip. JSON for page #{page} exists :: {json_per_page}")
                pages.append(json_per_page)
                continue

            response = self.download({"page": page})
            save(json_per_page, response)

            pages.append(json_per_page)

            if self.verbose:
                print(f"✅ JSON for page #{page} downloaded :: {json_per_page}")
                print("=" * 60)

        self.files = self.files + pages
        return pages

    def download_first_page(self) -> dict:
        """Download first page to be used as reference and get metadata pagination

        Returns:
            response (dict): Dictionary of response
        """
        first_page_json = os.path.join(self.page_dir, f"{self.json_filename}_1.json")

        if os.path.isfile(first_page_json):
            if self.verbose:
                print(f"✅ JSON First Page exists :: {first_page_json}")
            self.files.append(first_page_json)
            return json.load(open(first_page_json))

        if self.verbose:
            print(f"⌛ Downloading JSON First Page :: {first_page_json}")

        response = self.download()

        # Saving first JSON
        save(first_page_json, response)

        if self.verbose:
            total = response["total"]
            print(f"ℹ️ Total Data :: {total}")
            print(f"✅ JSON First Page downloaded :: {first_page_json}")

        self.files.append(first_page_json)

        return response

    def extract(self) -> Self:
        """Extract description

        Returns:
            self: Self
        """
        self.df = extract(
            earthquakes=self.events, locale=self.locale, verbose=self.verbose
        )
        return self

    def transform(self) -> Self:
        """Transform all json files

        Returns:
            self: Self
        """
        self.events = transform(files=self.files)
        return self

    def var(self, as_daily: bool = False) -> Self:
        """Download VAR from Filter URL and save it to JSON files

        Args:
            as_daily (bool, optional): Whether to save daily json file. Defaults to False.

        Returns:
            self: Self
        """
        if as_daily:
            if self.verbose:
                print(f"ℹ️ Save VAR as daily JSON file.")
            self.download_daily(overwrite=True)
            return self

        # Creating pages directory
        os.makedirs(self.page_dir, exist_ok=True)

        response = self.download_first_page()
        self.download_per_page(response)

        if self.verbose:
            print("ℹ️ JSON Files :: ", len(self.files))

        self.transform().extract()

        if len(self.events) != len(self.df):
            if self.verbose:
                print(
                    f"⚠️ Events length not equal to DataFrame length :: {len(self.events)} vs {len(self.df)}"
                )
            print(f"⚠️ Please kindly to check the results.")

        return self

    def _to(
        self, filetype: str, filename: str = None, verbose: bool = True
    ) -> str | None:
        """Save wrapper

        Args:
            filetype (str): Type of wrapper. 'csv' or 'xlsx'
            filename (str, optional): Name of file. Defaults to None.
            verbose (bool, optional): Print verbose output. Defaults to True.

        Returns:
            Saved path location
        """
        if len(self.df) == 0:
            print(f"⚠️ No data found.")
            return None

        assert filetype in ["csv", "xlsx"], TypeError(
            f"❌ {filetype} not supported. Choose 'csv' or 'xlsx'."
        )

        to_dir = os.path.join(self.magma_dir, filetype)
        os.makedirs(to_dir, exist_ok=True)

        _filename = filename if filename is not None else self.filename
        path = os.path.join(to_dir, f"{_filename}.{filetype}")

        (
            self.df.to_csv(path, index=False)
            if filetype == "csv"
            else self.df.to_excel(path, index=False)
        )

        if verbose:
            print(f"🗃️ Saved to :: {path}")

        return path

    def to_csv(self, filename: str = None, verbose: bool = True) -> str | None:
        """Save to CSV files

        Args:
            filename (str, optional): CSV filename. Defaults to None.
            verbose (bool, optional): For debugging. Defaults to True.

        Returns:
            str: CSV file location
        """
        return self._to(filetype="csv", filename=filename, verbose=verbose)

    def to_excel(self, filename: str = None, verbose: bool = True) -> str | None:
        """Save to Excel Files

        Args:
            filename (str, optional): Excel filename. Defaults to None.
            verbose (bool, optional): For debugging. Defaults to True.
        """
        return self._to(filetype="xlsx", filename=filename, verbose=verbose)



import os
from datetime import datetime

from magma_var.utils import check_directory


class MagmaVar:
    def __init__(
        self,
        volcano_code: str,
        start_date: str,
        end_date: str,
        current_dir: str = None,
        locale: str = "id",
        verbose: bool = False,
        debug: bool = False,
    ):
        self.volcano_code = volcano_code.upper()
        self.start_date = start_date
        self.end_date = end_date
        self.locale = locale

        if current_dir is None:
            current_dir = os.getcwd()
        self.current_dir = current_dir

        self._validate()

        self.output_dir, self.figures_dir, self.magma_dir = check_directory(current_dir)
        self.magma_dir = os.path.join(self.magma_dir, "var")
        os.makedirs(self.magma_dir, exist_ok=True)

        self.json_dir: str = os.path.join(self.magma_dir, "json")

        self.filename = f"{self.volcano_code}_{self.start_date}_{self.end_date}"

        self.verbose = verbose
        self.debug = debug

    def _validate(self) -> None:
        """Validating parameters.

        Returns:
            None

        Raises:
            ValueError: If parameters are invalid.
            IsADirectoryError: If directory does not exist.
        """
        assert datetime.fromisoformat(self.start_date), ValueError(
            "❌ Start date must be in ISO 8601 format (YYYY-MM-DD)"
        )

        assert datetime.fromisoformat(self.end_date), ValueError(
            "❌ End date must be in ISO 8601 format (YYYY-MM-DD)"
        )

        assert datetime.fromisoformat(self.start_date) <= datetime.fromisoformat(
            self.end_date
        ), ValueError("❌ Start date must be before end date")

        assert self.locale in ["id", "en"], ValueError(
            f"❌ Locale must be 'id' or 'en'"
        )

        assert os.path.isdir(self.current_dir), IsADirectoryError(
            f"❌ Current directory: {self.current_dir} is not a directory"
        )



from dataclasses import dataclass
from functools import cached_property
from typing import List, Dict, Optional

import pandas as pd
from magma_var.const import EARTHQUAKE_TYPES


@dataclass
class Earthquake:
    """Earthquake metadata.

    Attributes:
        code (str): Earthquake code
        alias (str): Earthquake alias
        earthquake_id (str): Earthquake name in Indonesia
        earthquake_en (str): Earthquake name in English
        color (str): Earthquake color in hex
        column_name (str): Earthquake column name
    """

    code: str
    alias: str
    earthquake_id: str
    earthquake_en: str
    color: str
    column_name: str


@dataclass
class Earthquakes:
    """List of Earthquake.

    Attributes:
        earthquakes (List[Earthquake]): List of Earthquake class
    """

    earthquakes: List[Earthquake]

    def __iter__(self):
        return iter(self.earthquakes)

    def __len__(self):
        return len(self.earthquakes)

    @property
    def df(self) -> pd.DataFrame:
        """DataFrame representation."""
        return pd.DataFrame(self.earthquakes)

    @property
    def codes(self) -> List[str]:
        """List of Earthquake codes."""
        return self.df["code"].to_list()

    @property
    def colors(self) -> List[str]:
        """List of Earthquake codes."""
        return self.df["color"].to_list()

    def columns(self, locale: Optional[str] = "id") -> Dict[str, str]:
        """List of Earthquake column names.

        Args:
            locale (str, optional): Earthquake column name. Defaults to "id".

        Returns:
            Dict[str, str]: List of Earthquake column names.
        """
        columns = {}
        _dict = self.df[["column_name", f"earthquake_{locale}"]].to_dict("records")
        for row in _dict:
            columns[row["column_name"]] = row[f"earthquake_{locale}"]
        return columns

    def where(self, attribute: str, value: str) -> Earthquake | None:
        """Query earthquake by attribute and value.

        Args:
            attribute (str): Earthquake attribute
            value (str): Earthquake value

        Returns:
            Earthquake | None: Earthquake class if found
        """
        for earthquake in self.earthquakes:
            if getattr(earthquake, attribute) == value:
                return earthquake
        return None


earthquakes: Earthquakes = Earthquakes(
    earthquakes=[Earthquake(**eq) for eq in EARTHQUAKE_TYPES]
)
